#include "config_private.h"
#include "iocall.h"
#include "ctx.h"
#include "stdlib.h"
#include "objects.h"
#include "vm.h"

static const wt_u32 iocall_retv___none[] = { 0 };
static const wt_u32 iocall_retv_u32[] = { WTRT_TYPEINFO(0, WTTYPE_U32) };
static const wt_u32 iocall_retv_i64[] = { WTRT_TYPEINFO(0, WTTYPE_I64) };
static const wt_u32 iocall_retv_bool[] = { WTRT_TYPEINFO(0, WTTYPE_BOOL) };
static const wt_u32 iocall_retv_bool_char_u64_u64_u64[] = {
	WTRT_TYPEINFO(0, WTTYPE_BOOL), WTRT_TYPEINFO(0, WTTYPE_CHAR),
	WTRT_TYPEINFO(0, WTTYPE_U64), WTRT_TYPEINFO(0, WTTYPE_U64),
	WTRT_TYPEINFO(0, WTTYPE_U64) };
static const wt_u32 iocall_retv_file[] = { WTRT_TYPEINFO(0, WTTYPE_FILE) };
static const wt_u32 iocall_retv_dir[] = { WTRT_TYPEINFO(0, WTTYPE_DIR) };
static const wt_u32 iocall_retv_char[] = { WTRT_TYPEINFO(0, WTTYPE_CHAR) };

static enum wtr_runresult iocall_make_string_from_return(
	struct wtr_context* const ctx)
{
	struct wtr_vm* const      vm   = ctx->vm;
	struct wtr_program* const prog = ctx->frame.prog;

	wt_u32              len;
	wt_vmptr            vma;
	struct wtr_objentry ent;

	/* TODO: this is a bit negligent */
	if(WTR_CTX_DS_SIZE(ctx)<sizeof(wt_u32) ||
		WTR_CTX_DS_SPACE(ctx)<sizeof(wt_vmptr))
	{
		ctx->last_exception = StackOverflowException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_POP(ctx, wt_u32, len);

	if(0==(vma=wtr_str_create_n(ctx->vm, (char*)ctx->iocall_npriv, len)))
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_vmptr, vma);

	ent.obj     = vma;
	ent.pfn.ptr = prog;
	ent.pfn.ext = ctx->vm->pc_strd_leave;

	if(NULL==wtr_dynstack_push_p(&ctx->tos, &ent))
	{
		wtr_str_delete(ctx->vm, vma);
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

void wtr_iocall_writestdout(struct wtr_context* const ctx, wt_vmptr vma,
	wt_u32 len)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_WRITESTDOUT;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 0;
	ctx->iocall_retv = iocall_retv___none;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_STRING);
	arg->v_string.p    = vma;
	arg->v_string.n    = len;

	ctx->iocall_next = NULL;
}

void wtr_iocall_writestdout_x(struct wtr_context* const ctx, const char* str,
	wt_u32 len)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_WRITESTDOUT;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 0;
	ctx->iocall_retv = iocall_retv___none;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_XSTRING);
	arg->v_xstring.p   = (char*)str;
	arg->v_xstring.n   = len;

	ctx->iocall_next = NULL;
}

void wtr_iocall_readlnstdin(struct wtr_context* const ctx, char sep)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_READLNSTDIN;

	ctx->iocall_argc = 2;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_u32;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_XSTRING);
	arg->v_xstring.p   = ctx->sbuf;
	arg->v_xstring.n   = ctx->sbuf_size - 1;
	++arg;
	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_CHAR);
	arg->v_char        = (wt_u8)sep;

	ctx->iocall_npriv = ctx->sbuf;
	ctx->iocall_next  = iocall_make_string_from_return;
}

void wtr_iocall_access(struct wtr_context* const ctx, wt_vmptr pathname,
	wt_u32 namelen, wt_u32 mode)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_ACCESS;

	ctx->iocall_argc = 2;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_bool;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_STRING);
	arg->v_string.p    = pathname;
	arg->v_string.n    = namelen;
	++arg;
	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_U32);
	arg->v_u32         = mode;

	ctx->iocall_next = NULL;
}

void wtr_iocall_stat(struct wtr_context* const ctx, wt_vmptr pathname,
	wt_u32 namelen)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_STAT;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 5;
	ctx->iocall_retv = iocall_retv_bool_char_u64_u64_u64;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_STRING);
	arg->v_string.p    = pathname;
	arg->v_string.n    = namelen;

	ctx->iocall_next = NULL;
}

void wtr_iocall_fstat(struct wtr_context* const ctx, wt_file file)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FSTAT;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 5;
	ctx->iocall_retv = iocall_retv_bool_char_u64_u64_u64;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_FILE);
	arg->v_file        = file;

	ctx->iocall_next = NULL;
}

void wtr_iocall_fopen(struct wtr_context* const ctx, wt_vmptr pathname,
	wt_u32 namelen, char mode)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FOPEN;

	ctx->iocall_argc = 2;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_file;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_STRING);
	arg->v_string.p    = pathname;
	arg->v_string.n    = namelen;
	++arg;
	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_CHAR);
	arg->v_u32         = mode;

	ctx->iocall_next = NULL;
}

void wtr_iocall_fclose(struct wtr_context* const ctx, wt_file file)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FCLOSE;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 0;
	ctx->iocall_retv = iocall_retv___none;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_FILE);
	arg->v_file        = file;

	ctx->iocall_next = NULL;
}

void wtr_iocall_fseek(struct wtr_context* const ctx, wt_file file,
	wt_i64 offset, wt_u32 whence)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FSEEK;

	ctx->iocall_argc = 3;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_i64;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_FILE);
	arg->v_file        = file;
	++arg;
	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_I64);
	arg->v_i64         = offset;
	++arg;
	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_U32);
	arg->v_u32         = whence;

	ctx->iocall_next = NULL;
}

void wtr_iocall_feof(struct wtr_context* const ctx, wt_file file)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FEOF;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_bool;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_FILE);
	arg->v_file        = file;

	ctx->iocall_next = NULL;
}

static enum wtr_runresult iocall_fwrite_checkreturn(
	struct wtr_context* const ctx)
{
	struct wtr_vm* const vm = ctx->vm;

	wt_u32 act_len;

	if(WTR_CTX_DS_SIZE(ctx)<sizeof(wt_u32))
	{
		ctx->last_exception = StackOverflowException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_POP(ctx, wt_u32, act_len);

	if(act_len!=(wt_uintptr)ctx->iocall_npriv)
	{
		ctx->last_exception = IOException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

static enum wtr_runresult iocall_fwriteln_followup(
	struct wtr_context* const ctx)
{
	const enum wtr_runresult rr = iocall_fwrite_checkreturn(ctx);

	if(WTRR_RUN!=rr || 0==ctx->iocall_wlen)
	{
		return rr;
	}

	wtr_iocall_fwrite_x(ctx, ctx->iocall_argv[0].v_file,
		(char*)ctx->iocall_wbuf, ctx->iocall_wlen, 1);

	/* clear the 'wbuf' */
	ctx->iocall_wlen = 0;

	return WTRR_IOCALL;
}

void wtr_iocall_fwrite(struct wtr_context* const ctx, wt_file file,
	wt_vmptr vma, wt_u32 len, int pop_and_check_return_value)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FWRITE;

	ctx->iocall_argc = 2;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_u32;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_FILE);
	arg->v_file        = file;
	++arg;
	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_STRING);
	arg->v_string.p    = vma;
	arg->v_string.n    = len;

	if(pop_and_check_return_value)
	{
		ctx->iocall_npriv = (void*)(wt_uintptr)len;
		ctx->iocall_next  = iocall_fwrite_checkreturn;
	}
	else
	{
		ctx->iocall_next = NULL;
	}
}

void wtr_iocall_fwrite_x(struct wtr_context* const ctx, wt_file file,
	const char* buf, wt_u32 len, int pop_and_check_return_value)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FWRITE;

	ctx->iocall_argc = 2;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_u32;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_FILE);
	arg->v_file        = file;
	++arg;
	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_XSTRING);
	arg->v_xstring.p   = (char*)buf;
	arg->v_xstring.n   = len;

	if(pop_and_check_return_value)
	{
		ctx->iocall_npriv = (void*)(wt_uintptr)len;
		ctx->iocall_next  = iocall_fwrite_checkreturn;
	}
	else
	{
		ctx->iocall_next = NULL;
	}
}

void wtr_iocall_fwriteln(struct wtr_context* const ctx, wt_file file,
	wt_vmptr vma, wt_u32 len, char sep)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FWRITE;

	ctx->iocall_argc = 2;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_u32;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_FILE);
	arg->v_file        = file;
	++arg;
	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_STRING);
	arg->v_string.p    = vma;
	arg->v_string.n    = len;

	ctx->iocall_wlen    = 1;
	ctx->iocall_wbuf[0] = sep;

	ctx->iocall_npriv = (void*)(wt_uintptr)len;
	ctx->iocall_next  = iocall_fwriteln_followup;
}

static enum wtr_runresult iocall_fread_checkreturn(
	struct wtr_context* const ctx)
{
	struct wtr_vm* const vm = ctx->vm;

	wt_u32 act_len;

	if(WTR_CTX_DS_SIZE(ctx)<sizeof(wt_u32))
	{
		ctx->last_exception = StackOverflowException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_POP(ctx, wt_u32, act_len);

	if(act_len!=(wt_uintptr)ctx->iocall_npriv)
	{
		ctx->last_exception = EndOfFileException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

void wtr_iocall_fread(struct wtr_context* const ctx, wt_file file,
	wt_vmptr vma, wt_u32 maxlen, int pop_and_check_return_value)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FREAD;

	ctx->iocall_argc = 2;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_u32;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_FILE);
	arg->v_file        = file;
	++arg;
	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_STRING);
	arg->v_string.p    = vma;
	arg->v_string.n    = maxlen;

	if(pop_and_check_return_value)
	{
		ctx->iocall_npriv = (void*)(wt_uintptr)maxlen;
		ctx->iocall_next  = iocall_fread_checkreturn;
	}
	else
	{
		ctx->iocall_next = NULL;
	}
}

void wtr_iocall_freadln(struct wtr_context* const ctx, wt_file file,
	char* buf, wt_u32 maxlen, char sep)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FREADLN;

	ctx->iocall_argc = 3;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_u32;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_FILE);
	arg->v_file        = file;
	++arg;
	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_XSTRING);
	arg->v_xstring.p   = buf;
	arg->v_xstring.n   = maxlen;
	++arg;
	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_CHAR);
	arg->v_char        = sep;

	ctx->iocall_npriv = buf;
	ctx->iocall_next  = iocall_make_string_from_return;
}

void wtr_iocall_opendir(struct wtr_context* const ctx, wt_vmptr pathname,
	wt_u32 namelen)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_OPENDIR;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_dir;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_STRING);
	arg->v_string.p    = pathname;
	arg->v_string.n    = namelen;

	ctx->iocall_next = NULL;
}

void wtr_iocall_closedir(struct wtr_context* const ctx, wt_dir dir)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_CLOSEDIR;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 0;
	ctx->iocall_retv = iocall_retv___none;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_DIR);
	arg->v_dir         = dir;

	ctx->iocall_next = NULL;
}

void wtr_iocall_readdir(struct wtr_context* const ctx, wt_dir dir)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_READDIR;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_bool;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_DIR);
	arg->v_dir         = dir;

	ctx->iocall_next = NULL;
}

void wtr_iocall_dirgettype(struct wtr_context* const ctx, wt_dir dir)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_DIRGETTYPE;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_char;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_DIR);
	arg->v_dir         = dir;

	ctx->iocall_next = NULL;
}

void wtr_iocall_dirgetname(struct wtr_context* const ctx, wt_dir dir,
	char* buf, wt_u32 maxlen)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_DIRGETNAME;

	ctx->iocall_argc = 2;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_u32;

	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_DIR);
	arg->v_dir         = dir;
	++arg;
	arg->writable_type = WTRT_TYPEINFO(0, WTTYPE_XSTRING);
	arg->v_xstring.p   = buf;
	arg->v_xstring.n   = maxlen;

	ctx->iocall_npriv = buf;
	ctx->iocall_next  = iocall_make_string_from_return;
}
