#include "config_private.h"
#include "sym3.h"
#include "sym.h"
#include "lreader.h"

/* returns offset after terminating zero (or 0 on error); 'bufsz' must */
/* include space for the terminating zero */
static wt_uint32_t read_name(struct wtr_lreader* const lr, wt_uint32_t offs,
	char* buf, size_t bufsz)
{
	while(0<bufsz)
	{
		--bufsz;

		if(0==((*buf++)=lr->read8(lr, offs++)))
		{
			return offs;
		}
	}

	return 0;
}

/* translate WTB3's type representation to runtime type representation */
static int wtr_sym3_translate_typeid(wt_typeid* const out, wt_uint32_t type3)
{
	wt_typeid t;
	wt_typeid f = 0;

	if(WTB3TM_UNUSEDBITS & type3)
	{
		return 1;
	}

	if(WTB3TF_LOCAL & type3)
	{
		t = WTB3TM_USERTYPEID & type3;
		f |= WTRTF_USER;
	}
	else
	{
		switch(WTB3TM_BASICTYPE & type3)
		{
		case WTB3T_BOOL:   t = WTTYPE_BOOL;   break;
		case WTB3T_I32:    t = WTTYPE_I32;    break;
		case WTB3T_U32:    t = WTTYPE_U32;    break;
		case WTB3T_I64:    t = WTTYPE_I64;    break;
		case WTB3T_U64:    t = WTTYPE_U64;    break;
		case WTB3T_F32:    t = WTTYPE_F32;    break;
		case WTB3T_F64:    t = WTTYPE_F64;    break;
		case WTB3T_CHAR:   t = WTTYPE_CHAR;   break;
		case WTB3T_STRING: t = WTTYPE_STRING; break;
		case WTB3T_HANDLE: t = WTTYPE_HANDLE; break;
		case WTB3T_FILE:   t = WTTYPE_FILE;   break;
		case WTB3T_DIR:    t = WTTYPE_DIR;    break;
		case WTB3T_I8:     t = WTTYPE_I8;     break;
		case WTB3T_U8:     t = WTTYPE_U8;     break;
		case WTB3T_I16:    t = WTTYPE_I16;    break;
		case WTB3T_U16:    t = WTTYPE_U16;    break;

		default:
			return 2;
		}
	}

	if(WTB3TF_REF     & type3) { f |= WTRTF_REF;     }
	if(WTB3TF_MUTABLE & type3) { f |= WTRTF_MUTABLE; }
	if(WTB3TF_SEQ     & type3) { f |= WTRTF_SEQ;     }

	*out = WTRT_TYPEINFO(f, t);
	return 0;
}

static wt_error wtr_sym3_read_funcsig(struct wtr_vm* const vm,
	struct wtr_lreader* const lr, struct wtr_s_funcsig* const funcsig,
	wt_uint32_t offs, wt_typeid* typs)
{
	wt_uint32_t cnt;

	/* argc */

	cnt = lr->read32(lr, offs);
	offs += sizeof(cnt);

	funcsig->argc = cnt;
	funcsig->argv = typs;

	while(0<(cnt--))
	{
		wt_uint32_t typeid = lr->read32(lr, offs);

		offs += sizeof(wt_uint32_t);

		if(0!=wtr_sym3_translate_typeid(typs, typeid))
		{
			return WTE_TYPEINF;
		}

		++typs;
	}

	/* retc */

	cnt = lr->read32(lr, offs);
	offs += sizeof(cnt);

	funcsig->retc = cnt;
	funcsig->retv = typs;

	while(0<(cnt--))
	{
		wt_uint32_t typeid = lr->read32(lr, offs);

		offs += sizeof(wt_uint32_t);

		if(0!=wtr_sym3_translate_typeid(typs, typeid))
		{
			return WTE_TYPEINF;
		}

		++typs;
	}

	/* pure */

	funcsig->is_pure = lr->read8(lr, offs);

	return WTE_OK;
}

wt_error wtr_sym3_parse(struct wtr_vm* const vm,
	struct wtr_lreader* const lr, struct wtr_symtab* const symtab,
	struct wtr_sym** const p_sym, const wt_uint32_t shoffs,
	struct wtb3_sym* const symhdr, const enum wtr_symbind symbind)
{
	char name[WTR_MAX_SYMNAME_LEN+1];
	char buf[WTR_MAX_SYMNAME_LEN+1];
	wt_uint32_t offs;
	wt_uint32_t namelen;
	wt_uint32_t symsize;
	wt_uint32_t strsize;
	wt_uint32_t aldatasize;
	wt_uint32_t* aldata; /* aligned data */
	char* str;
	enum wtr_symtype symtype;
	struct wtr_sym* sym;

	/* fetch symbol name */
	{
		const wt_uint32_t start_offs = shoffs + symhdr->sym_name;

		if(0==(offs=read_name(lr, start_offs, name, sizeof(name))))
		{
			return WTE_SYMNAME;
		}

		/* symbol name must be at least 1 character long! */

		if(1>=(offs - start_offs))
		{
			return WTE_SYMNAME;
		}

		namelen = (offs - start_offs) - 1;
	}

	symsize = 0;
	strsize = 0;
	aldatasize = 0;

	offs = shoffs + symhdr->sym_offs;

	switch(symhdr->sym_type)
	{
	case WTB3SYMT_EXCEPTION:
		symtype = WTR_SYMTYPE_EXC;
		symsize = sizeof(struct wtr_sym_exc);
		break;

	/* NOTE: WTB3SYMT_TYPE is defined, but not used (yet?) */
/*
	case WTB3SYMT_TYPE:
		symtype = WTR_SYMTYPE_TYPEAL;
		symsize = sizeof(struct wtr_sym_type);
		break;
*/

	case WTB3SYMT_OPERATOR:
		symtype = WTR_SYMTYPE_OPER;
		symsize = sizeof(struct wtr_sym_oper);
		break;

	case WTB3SYMT_PRIVATERECORD:
		symtype = WTR_SYMTYPE_PRIVREC;
		symsize = sizeof(struct wtr_sym_privrec);
		break;

	case WTB3SYMT_LIMITEDRECORD:
		symtype = WTR_SYMTYPE_LIMREC;
		symsize = sizeof(struct wtr_sym_limrec);
		break;

	case WTB3SYMT_RECORD:
		symtype = WTR_SYMTYPE_REC;
		symsize = sizeof(struct wtr_sym_rec);
		{
			wt_uint32_t fldc = lr->read32(lr, offs);
			wt_uint32_t start;

			aldatasize = fldc * sizeof(struct wtr_s_recfield);
			offs += sizeof(fldc) + fldc*sizeof(wt_uint32_t);
			start = offs;

			while(0<(fldc--))
			{
				if(0==(offs=read_name(lr, offs, buf, sizeof(buf))))
				{
					return WTE_SYMNAME;
				}
			}

			strsize = offs - start;
		}
		break;

	case WTB3SYMT_ENUM:
		symtype = WTR_SYMTYPE_ENUM;
		symsize = sizeof(struct wtr_sym_enum);
		{
			wt_uint32_t enumc = lr->read32(lr, offs);
			wt_uint32_t start;

			aldatasize = enumc * sizeof(char*);
			offs += sizeof(enumc);
			start = offs;

			while(0<(enumc--))
			{
				if(0==(offs=read_name(lr, offs, buf, sizeof(buf))))
				{
					return WTE_SYMNAME;
				}
			}

			strsize = offs - start;
		}
		break;

	case WTB3SYMT_DATATYPE:
		symtype = WTR_SYMTYPE_DATATYPE;
		symsize = sizeof(struct wtr_sym_datatype);
		{
			wt_uint32_t numcons = lr->read32(lr, offs);

			aldatasize = numcons * sizeof(struct wtr_s_datacons);
			offs += sizeof(numcons) * 2; /* skip defcons! */

			while(0<(numcons--))
			{
				wt_uint32_t start = offs;
				wt_uint32_t typn;

				if(0==(offs=read_name(lr, offs, buf, sizeof(buf))))
				{
					return WTE_SYMNAME;
				}

				strsize += offs - start;

				offs = (offs + 3) & (-4);

				typn = lr->read32(lr, offs);
				aldatasize += typn * sizeof(wt_uint32_t);

				offs += sizeof(typn) + typn*sizeof(wt_uint32_t);
			}
		}
		break;

	case WTB3SYMT_FUNCTIONTYPE:
		symtype = WTR_SYMTYPE_FUNCTYPE;
		symsize = sizeof(struct wtr_sym_functype);
		{
			goto compute_funcsig_size;
		}
		break;

	case WTB3SYMT_FUNCTION:
		symtype = WTR_SYMTYPE_WTFUNC;
		symsize = sizeof(struct wtr_sym_wtfunc);
	compute_funcsig_size:;
		{
			wt_uint32_t cnt;

			/* argc */
			cnt = lr->read32(lr, offs);
			aldatasize += cnt * sizeof(wt_typeid);
			offs += sizeof(cnt) + cnt*sizeof(wt_uint32_t);

			/* retv */
			cnt = lr->read32(lr, offs);
			aldatasize += cnt * sizeof(wt_typeid);
		}
		break;

	default:
		return WTE_BINERR;
	}

	if(NULL==(sym=wtr_sym_new(vm, symtype, symbind, symtab,
		symsize+aldatasize+strsize, name, namelen)))
	{
		return WTE_NOMEM;
	}

	aldata = (wt_uint32_t*)(((char*)sym) + symsize);
	str    = ((char*)aldata) + aldatasize;
	offs   = shoffs + symhdr->sym_offs;

	switch(symhdr->sym_type)
	{
	case WTB3SYMT_EXCEPTION:
		break;

	case WTB3SYMT_OPERATOR:
		{
			struct wtr_sym_oper* const asym = (struct wtr_sym_oper*)sym;

			wt_uint32_t prec  = WTB3SYM_OPPREC(symhdr->sym_info);
			wt_uint32_t assoc = WTB3SYM_OPASSOC(symhdr->sym_info);

			asym->oper_id = WTR_SYM_UNDEF;

			/* check precedence against binary spec */

			if(WTB3SYM_OP_MINPREC>prec || WTB3SYM_OP_MAXPREC<prec)
			{
				goto err_badsym;
			}

			asym->precedence = prec;

			/* check and translate associativity */

			switch(assoc)
			{
			case WTB3SYM_OP_LEFTASSOC:
				asym->associativity = WTR_SYM_OP_ASSOC_LEFT;
				break;

			case WTB3SYM_OP_RIGHTASSOC:
				asym->associativity = WTR_SYM_OP_ASSOC_RIGHT;
				break;

			default:
				goto err_badsym;
			}
		}
		break;

/*
	case WTB3SYMT_TYPE:
		{
			struct wtr_sym_type* const asym = (struct wtr_sym_type*)sym;
			wt_uint32_t t = lr->read32(lr, offs);
			wtr_sym3_translate_typeid(&typeid, t);
		}
		break;
*/

	case WTB3SYMT_PRIVATERECORD:
	case WTB3SYMT_LIMITEDRECORD:
		{
			struct wtr_sym_type* const asym = (struct wtr_sym_type*)sym;

			asym->typeinfo = WTR_SYM_UNDEF;
		}
		break;

	case WTB3SYMT_RECORD:
		{
			struct wtr_sym_rec* const asym = (struct wtr_sym_rec*)sym;

			wt_uint32_t fldc = lr->read32(lr, offs);
			wt_uint32_t i;

			struct wtr_s_recfield* const fldv =
				(struct wtr_s_recfield*)aldata;

			asym->type.typeinfo = WTR_SYM_UNDEF;

			offs += sizeof(fldc);

			asym->fldc = fldc;
			asym->fldv = fldv;

			for(i=0;i<fldc;++i)
			{
				wt_uint32_t typeid = lr->read32(lr, offs);

				if(0!=wtr_sym3_translate_typeid(&fldv[i].type, typeid))
				{
					goto err_typeinf;
				}

				offs += sizeof(wt_uint32_t);
			}

			for(i=0;i<fldc;++i)
			{
				const wt_uint32_t start = offs;

				if(0==(offs=read_name(lr, offs, str, sizeof(name))))
				{
					goto err_symname;
				}

				fldv[i].name = str;
				str += offs - start;
			}
		}
		break;

	case WTB3SYMT_ENUM:
		{
			struct wtr_sym_enum* const asym = (struct wtr_sym_enum*)sym;

			wt_uint32_t enumc = lr->read32(lr, offs);
			wt_uint32_t i;

			char** const enumv = (char**)aldata;

			asym->type.typeinfo = WTR_SYM_UNDEF;

			offs += sizeof(enumc);

			asym->enumeratorc = enumc;
			asym->enumeratorv = (const char**)enumv;

			for(i=0;i<enumc;++i)
			{
				const wt_uint32_t start = offs;

				if(0==(offs=read_name(lr, offs, str, sizeof(name))))
				{
					goto err_symname;
				}

				enumv[i] = str;
				str += offs - start;
			}
		}
		break;

	case WTB3SYMT_DATATYPE:
		{
			struct wtr_sym_datatype* const asym =
				(struct wtr_sym_datatype*)sym;

			wt_uint32_t numcons = lr->read32(lr, offs);
			wt_uint32_t i;

			struct wtr_s_datacons* const conses =
				(struct wtr_s_datacons*)aldata;

			wt_typeid* typs = (wt_typeid*)(conses + numcons);

			asym->type.typeinfo = WTR_SYM_UNDEF;

			offs += sizeof(numcons);

			asym->numcons = numcons;
			asym->defcons = lr->read32(lr, offs);
			asym->conses  = conses;

			offs += sizeof(wt_uint32_t); /* defcons */

			for(i=0;i<numcons;++i)
			{
				/* constructor name */
				{
					const wt_uint32_t start = offs;

					if(0==(offs=read_name(lr, offs, str, sizeof(name))))
					{
						goto err_symname;
					}

					offs = (offs + 3) & (-4);

					conses[i].name = str;
					str += offs - start;
				}

				/* constructor types */
				{
					wt_uint32_t typc = lr->read32(lr, offs);

					offs += sizeof(wt_uint32_t);

					conses[i].typc = typc;
					conses[i].typv = typs;

					while(0<(typc--))
					{
						wt_uint32_t typeid = lr->read32(lr, offs);

						offs += sizeof(wt_uint32_t);

						if(0!=wtr_sym3_translate_typeid(typs, typeid))
						{
							goto err_typeinf;
						}

						++typs;
					}
				}
			}
		}
		break;

	case WTB3SYMT_FUNCTION:
		{
			struct wtr_sym_wtfunc* const asym = (struct wtr_sym_wtfunc*)sym;

			wt_error err;

			asym->func.func_id = WTR_SYM_UNDEF;

			if(WTE_OK!=(err=wtr_sym3_read_funcsig(vm, lr, &asym->func.sig,
				offs, (wt_typeid*)aldata)))
			{
				wtr_sym_release(vm, sym);
				return err;
			}
		}
		break;

	case WTB3SYMT_FUNCTIONTYPE:
		{
			struct wtr_sym_functype* const asym =
				(struct wtr_sym_functype*)sym;

			wt_error err;

			asym->type.typeinfo = WTR_SYM_UNDEF;

			if(WTE_OK!=(err=wtr_sym3_read_funcsig(vm, lr, &asym->sig, offs,
				(wt_typeid*)aldata)))
			{
				wtr_sym_release(vm, sym);
				return err;
			}
		}
		break;

	default:
		break;
	}

	*p_sym = sym;

	return WTE_OK;

err_symname:
	wtr_sym_release(vm, sym);
	return WTE_SYMNAME;

err_typeinf:
	wtr_sym_release(vm, sym);
	return WTE_TYPEINF;

err_badsym:
	wtr_sym_release(vm, sym);
	return WTE_BADSYM;
}
