#include "config_private.h"
#include "vm.h"
#include "program.h"
#include "stdlib.h"
#include "byteorder.h"
#include "loader.h"
#include "processor.h"
#include "routines.h"
#include "vector.h"
#include "ctx.h"
#include "sym.h"
#include "symtab.h"
#include "symtab_b.h"
#include <libvmma/buddy32.h>
#include <string.h>
#include <stdlib.h>

/*----------------------------------------------------------------*/

wt_vmptr wtr_vm_vma_malloc(struct wtr_vm* const vm, size_t size)
{
	const wt_vmptr vma = buddy32v_alloc(vm->heap_buddy, size);

	if(0!=vma && 0!=wtr_vma_acquire(vm, vma, size))
	{
		buddy32v_free(vm->heap_buddy, vma);
		return 0;
	}

	return vma;
}

void wtr_vm_vma_free(struct wtr_vm* const vm, wt_vmptr memory)
{
	struct buddy32v* const buddy = vm->heap_buddy;

	const wt_u32 size = buddy32v_msize(buddy, memory);

	if(0==buddy32v_free(buddy, memory))
	{
		wtr_vma_release(vm, memory, size);
	}
}

int wtr_vma_acquire(struct wtr_vm* const vm, wt_vmptr vma, wt_u32 size)
{
	const wt_u32  szsh   = vm->vma_segsh;
	wt_u32* const segacq = vm->vma_segacq;

	wt_vmptr start_seg = vma >> szsh;
	wt_vmptr end_seg   = (vma + size - 1) >> szsh;

	for(;start_seg<=end_seg;++start_seg)
	{
		if(0==segacq[start_seg])
		{
			/* this function returns -VMMU_ERR_... on failure */
			if(0!=vmmu32_acquire_segment(&vm->vmmu,
				&vm->vmmu.segments[start_seg]))
			{
				/* TODO: rollback? */
				return 1;
			}
		}

		++segacq[start_seg];
	}

	return 0;
}

int wtr_vma_release(struct wtr_vm* const vm, wt_vmptr vma, wt_u32 size)
{
	const wt_u32  szsh   = vm->vma_segsh;
	wt_u32* const segacq = vm->vma_segacq;

	wt_vmptr start_seg = vma >> szsh;
	wt_vmptr end_seg   = (vma + size - 1) >> szsh;

	for(;start_seg<=end_seg;++start_seg)
	{
		--segacq[start_seg];

		if(0==segacq[start_seg])
		{
			/* TODO: we could check whether this failed or not */
			vmmu32_release_segment(&vm->vmmu,
				&vm->vmmu.segments[start_seg]);
		}
	}

	return 0;
}

void* wtr_vm_sys_realloc(struct wtr_vm* const vm, void* memory,
	const size_t old_size, const size_t new_size)
{
	void* const  alloc_ctx  = vm->alloc_ctx;
	WT_PFN_ALLOC alloc_func = vm->alloc_func;
	void*        new_mem;

	if(NULL==memory)
	{
		return alloc_func(alloc_ctx, NULL, new_size);
	}

	if(old_size==new_size)
	{
		/* this should not happen, but since they are equal do nothing */
		return memory;
	}

	if(NULL!=(new_mem=alloc_func(alloc_ctx, memory, new_size)))
	{
		return new_mem;
	}

	/* alloc_func() could not reallocate the block, but this could be */
	/* because it does not implement realloc, so try a different approach */

	if(NULL==(new_mem=alloc_func(alloc_ctx, NULL, new_size)))
	{
		/* nope, we sure ran out of memory */
		return NULL;
	}

	if(old_size>new_size)
	{
		memcpy(new_mem, memory, new_size);
	}
	else
	{
		memcpy(new_mem, memory, old_size);
	}

	alloc_func(alloc_ctx, memory, 0);

	return new_mem;
}

void* wtr_vm_sys_malloc(struct wtr_vm* const vm, size_t size)
{
	return vm->alloc_func(vm->alloc_ctx, NULL, size);
}

void wtr_vm_sys_free(struct wtr_vm* const vm, void* memory)
{
	vm->alloc_func(vm->alloc_ctx, memory, 0);
}

struct wtr_sym* wtr_vm_find_sym(struct wtr_vm* const vm,
	struct wtr_sym* const sym)
{
	struct wtr_list* const list = &vm->list_symtabs;
	struct wtr_node*       it;

	wtr_list_for_each(list, it)
	{
		struct wtr_symtab* const symtab =
			wtr_contof(it, struct wtr_symtab, link_symtabs);

		struct wtr_sym* const found = symtab->lookup(symtab, vm, sym);

		if(NULL!=found)
		{
			return found;
		}
	}

	return NULL;
}

size_t wtr_sizeof(struct wtr_vm* const vm, wt_typeid type)
{
	if(type & (WTRTF_SEQ | WTRTF_REF | WTRTF_USER))
	{
		/* TODO: user types could be simple type aliases */
		return sizeof(wt_vmptr);
	}

	if(WTRT_TYPEINFO(0, WTTYPE_XSTRING)==type)
	{
		return sizeof(wt_vmptr);
	}

	switch(WTRT_TYPEID(type))
	{
	case WTTYPE_BOOL:    return sizeof(wt_bool);
	case WTTYPE_STRING:  return sizeof(wt_vmptr);  /* this is an object */
	case WTTYPE_I32:     return sizeof(wt_i32);
	case WTTYPE_U32:     return sizeof(wt_u32);
	case WTTYPE_I64:     return sizeof(wt_i64);
	case WTTYPE_U64:     return sizeof(wt_u64);
#if WTRUNTIME_ENABLE_FLOAT32
	case WTTYPE_F32:     return sizeof(wt_f32);
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
#if WTRUNTIME_ENABLE_FLOAT64
	case WTTYPE_F64:     return sizeof(wt_f64);
#endif /* WTRUNTIME_ENABLE_FLOAT64 */
	case WTTYPE_CHAR:    return sizeof(wt_char);
	case WTTYPE_HANDLE:  return sizeof(wt_handle);
	case WTTYPE_FILE:    return sizeof(wt_handle); /* this is a pointer, but it will fit in this */
	case WTTYPE_DIR:     return sizeof(wt_handle); /* this is a pointer, but it will fit in this */

	case WTTYPE_I8:
	case WTTYPE_U8:
	case WTTYPE_I16:
	case WTTYPE_U16:
		return sizeof(wt_u32);
	}

	return 0;
}

wt_uint32_t wtr_vm_xorshift(struct wtr_vm* const vm)
{
	const wt_uint32_t t = vm->xs_x ^ (vm->xs_x << 11);
	vm->xs_x = vm->xs_y; vm->xs_y = vm->xs_z; vm->xs_z = vm->xs_w;
	return vm->xs_w = vm->xs_w ^ (vm->xs_w >> 19) ^ t ^ (t >> 8);
}

void wtr_vm_xorshift_seed(struct wtr_vm* const vm, wt_uint32_t seed)
{
	vm->xs_x = (seed);
	vm->xs_y = (seed >> 13);
	vm->xs_z = (seed << 7);
	vm->xs_w = (seed ^ 0xa7a7a7a7);

	wtr_vm_xorshift(vm);
}

/*---------------------------------------------------------------- */


static wt_vmptr wtr_vm_load_routine(struct wtr_vm* const vm, wt_vmptr vma,
	wt_uint32_t* const out_pc, const void* const data, const size_t size)
{
	*out_pc = vma / sizeof(wt_uint16_t);

	if(0!=wtr_vma_write(vm, vma, data, size))
	{
		return 0;
	}

	return vma + size;
}

static void release_vmmu_segments(struct wtr_vm* const vm)
{
	wt_u32 i;

	for(i=0;i<vm->vmmu.nsegments;++i)
	{
		vmmu32_release_segment(&vm->vmmu, &vm->vmmu.segments[i]);
	}
}

wt_error wtrCreateTombVM(struct wtr_vm** const p_vm, wt_u32 flags,
	struct vmmu32_segment* const vma_segs, size_t vma_nsegs, wt_u32 vma_szsh,
	wt_u32 heap_start, wt_u32 heap_size, wt_u32 heap_gran, wt_u8 ratio_css,
	wt_u8 ratio_dss, void* alloc_ctx, WT_PFN_ALLOC alloc_func)
{
	struct wtr_vm* vm        = NULL;
	wt_u32         acc_flags = 0; /* accepted flags (WTRVMF_*) */

	if(NULL==p_vm || 0==ratio_css || 0==ratio_dss || NULL==alloc_func)
	{
		return WTE_BADARG;
	}

	if(flags & WTRVMF_REQINTEGRITY)
	{
		acc_flags |= WTRVMF_REQINTEGRITY;
	}

	if(NULL==(vm=(struct wtr_vm*)alloc_func(alloc_ctx, NULL,
		sizeof(struct wtr_vm))))
	{
		return WTE_NOMEM;
	}

	memset(vm, 0, sizeof(struct wtr_vm));

	vm->next_prog_id   = 1;
	vm->alloc_ctx      = alloc_ctx;
	vm->alloc_func     = alloc_func;
	vm->ratio_css      = ratio_css;
	vm->ratio_dss      = ratio_dss;
	vm->accepted_flags = acc_flags;
	vm->vma_segsh      = vma_szsh;
	vm->xs_x           = (wt_u32)(wt_uintptr)vm;
	vm->xs_y           = (wt_u32)(wt_uintptr)wtrCreateTombVM;
	vm->xs_z           = 0x5a5a5a5a ^ vm->xs_x ^ vm->xs_y;
	vm->xs_w           = 0xa5a5a5a5 ^ vm->xs_x ^ vm->xs_y;

	wtr_list_init(&vm->list_programs);
	wtr_list_init(&vm->list_symtabs);

	{
		const int ec = -vmmu32_init(&vm->vmmu, vma_segs, vma_nsegs, vma_szsh);

		switch(ec)
		{
		case 0:
			/* success */
			break;

		case VMMU_ERR_EINVAL:
			return WTE_VINVAL;

		case VMMU_ERR_ENSEGS:
			return WTE_VNSEGS;

		case VMMU_ERR_ESEGSIZE:
			return WTE_VSEGSZ;

		default:
			return WTE_VINIT;
		}
	}

	if(NULL==(vm->vma_segacq=alloc_func(alloc_ctx, NULL,
		sizeof(vm->vma_segacq[0])*vma_nsegs)))
	{
		alloc_func(alloc_ctx, vm, 0);
		return WTE_NOMEM;
	}

	memset(vm->vma_segacq, 0, sizeof(vm->vma_segacq[0])*vma_nsegs);

	/* TODO: apply limits on 'heap_gran'? */
	{
		const size_t state_size = buddy32v_state_size(heap_size, heap_gran);

		if(0==state_size)
		{
			alloc_func(alloc_ctx, vm->vma_segacq, 0);
			alloc_func(alloc_ctx, vm, 0);
			return WTE_BADHEAP;
		}

		if(NULL==(vm->heap_sysmem=alloc_func(alloc_ctx, NULL, state_size)))
		{
			alloc_func(alloc_ctx, vm->vma_segacq, 0);
			alloc_func(alloc_ctx, vm, 0);
			return WTE_NOMEM;
		}

		memset(vm->heap_sysmem, 0, state_size);

		if(NULL==(vm->heap_buddy=buddy32v_init(vm->heap_sysmem, state_size,
			heap_start, heap_size, heap_gran)))
		{
			alloc_func(alloc_ctx, vm->vma_segacq, 0);
			alloc_func(alloc_ctx, vm->heap_sysmem, 0);
			alloc_func(alloc_ctx, vm, 0);
			return WTE_BADHEAP;
		}
	}

	if(0!=wtr_gidtree_create(&vm->gidtree, vm))
	{
		goto error_1;
	}

	if(NULL==(vm->symtab_b=wtr_symtab_b_create(vm)))
	{
		goto error_2;
	}

	{
		const size_t total_size = 0
			+ sizeof(wtr_routine_tcl_ja)
			+ sizeof(wtr_routine_tcl_leave)
			+ sizeof(wtr_routine_strd_leave)
			+ sizeof(wtr_routine_cfp_leave)
			;

		wt_vmptr vma = wtr_vm_vma_malloc(vm, total_size);

		if(0==vma)
		{
			goto error_3;
		}

		if(
			0==(vma=wtr_vm_load_routine(vm, vma, &vm->pc_tcl_ja,
				wtr_routine_tcl_ja, sizeof(wtr_routine_tcl_ja)))
			||
			0==(vma=wtr_vm_load_routine(vm, vma, &vm->pc_tcl_leave,
				wtr_routine_tcl_leave, sizeof(wtr_routine_tcl_leave)))
			||
			0==(vma=wtr_vm_load_routine(vm, vma, &vm->pc_strd_leave,
				wtr_routine_strd_leave, sizeof(wtr_routine_strd_leave)))
			||
			0==(vma=wtr_vm_load_routine(vm, vma, &vm->pc_cfp_leave,
				wtr_routine_cfp_leave, sizeof(wtr_routine_cfp_leave)))
		)
		{
			goto error_3;
		}
	}

	vm->external_tag = ~((wt_u32)0);

	*p_vm = vm;
	return WTE_OK;

error_3:;
	{
		struct wtr_symtab* const symtab = (struct wtr_symtab*)vm->symtab_b;
		symtab->destroy(symtab, vm);
	}
error_2:;
	wtr_gidtree_release(&vm->gidtree, vm);
error_1:
	release_vmmu_segments(vm);
	alloc_func(alloc_ctx, vm->vma_segacq, 0);
	alloc_func(alloc_ctx, vm->heap_sysmem, 0);
	alloc_func(alloc_ctx, vm, 0);
	return WTE_NOMEM;
}

wt_error wtrVMDestroy(struct wtr_vm* const vm)
{
	WT_PFN_ALLOC alloc_func = vm->alloc_func;
	void*        alloc_ctx  = vm->alloc_ctx;

	if(NULL==vm)
	{
		return WTE_BADARG;
	}

	if(0<vm->num_context)
	{
		return WTE_REFCTX;
	}

	{
		size_t num_symtabs = 0;

		struct wtr_list* const list = &vm->list_symtabs;
		struct wtr_node*       it;

		wtr_list_for_each(list, it)
		{
			++num_symtabs;
		}

		/* we expect 1 symtab only: the builtin/user symtab */

		if(1<num_symtabs)
		{
			return WTE_REFER;
		}
	}

	/* destroy the symtab */
	{
		struct wtr_symtab* const symtab = (struct wtr_symtab*)vm->symtab_b;
		symtab->destroy(symtab, vm);
	}

	wtr_gidtree_release(&vm->gidtree, vm);

	release_vmmu_segments(vm);

	alloc_func = vm->alloc_func;
	alloc_ctx  = vm->alloc_ctx;

	alloc_func(alloc_ctx, vm->vma_segacq, 0);
	alloc_func(alloc_ctx, vm->heap_sysmem, 0);
	alloc_func(alloc_ctx, vm, 0);

	return WTE_OK;
}

wt_error wtrVMSeedRNG(struct wtr_vm* const vm, wt_uint32_t seed)
{
	wtr_vm_xorshift_seed(vm, seed);
	return WTE_OK;
}

wt_error wtrVMSetLoaderOps(struct wtr_vm* const vm, void* const ops_ctx,
	const struct wtr_loaderops* const ops)
{
	vm->loaderops_ctx = ops_ctx;
	vm->loaderops     = ops;

	return WTE_OK;
}

wt_error wtrVMSetHashAlgs(struct wtr_vm* const vm,
	const struct wtr_hashalg* const algs)
{
	vm->hashalgs = algs;

	return WTE_OK;
}

static int test_exc_name(void* const ctx, struct wtr_sym* const sym)
{
	const enum wtr_symtype symtype = sym->type;

	if(!wtr_sym_is_exc(symtype))
	{
		return 1;
	}

	return strcmp((const char*)ctx, sym->name);
}

wt_error wtrVMGetException(struct wtr_vm* const vm, const char* const name,
	wt_exception* const p_exc)
{
	if(NULL==name || NULL==p_exc)
	{
		return WTE_BADARG;
	}
	else
	{
		wt_globalid gid;

		if(0!=wtr_gidtree_find(&vm->gidtree, &gid, (void*)name,
			test_exc_name))
		{
			return WTE_NOENT;
		}

		*p_exc = gid;
	}

	return WTE_OK;
}

wt_error wtrVMExceptionToString(struct wtr_vm* const vm, wt_exception exc,
	char* namebuf, size_t bufsize)
{
	if(NULL==namebuf)
	{
		return WTE_BADARG;
	}
	else
	{
		struct wtr_sym* const sym = wtr_gidtree_get(&vm->gidtree, exc);

		size_t len;

		if(NULL==sym || !wtr_sym_is_exc(sym->type))
		{
			return WTE_NOENT;
		}

		len = strlen(sym->name);

		if(len>=bufsize)
		{
			return WTE_NOFIT;
		}

		memcpy(namebuf, sym->name, len+1);
	}

	return WTE_OK;
}

wt_error wtrVMAddException(struct wtr_vm* const vm, const char* const name,
	wt_exception* const out_exc)
{
	if(NULL==name)
	{
		return WTE_BADARG;
	}
	else
	{
		wt_error err;

		struct wtr_sym_exc* const exc = wtr_symtab_b_add_exc(vm, vm->symtab_b,
			name, &err);

		if(NULL==exc)
		{
			return err;
		}

		if(NULL!=out_exc)
		{
			*out_exc = exc->exc_id;
		}
	}

	return WTE_OK;
}

wt_error wtrVMAddFunction(struct wtr_vm* const vm, const char* name,
	wt_uint32_t argc, const wt_typeid* argv, wt_uint32_t retc,
	const wt_typeid* retv, WT_PFN_EXTFUNC func, void* token)
{
	if(NULL==name || NULL==func || !(argc==0 || argv) || !(retc==0 || retv))
	{
		return WTE_BADARG;
	}
	else
	{
		wt_error err;

		struct wtr_sym_cfunc* const cfunc = wtr_symtab_b_add_func(vm,
			vm->symtab_b, name, argc, argv, retc, retv, func, token, &err);

		if(NULL==cfunc)
		{
			return err;
		}
	}

	return WTE_OK;
}

wt_error wtrVMGetFunction(struct wtr_vm* const vm,
	struct wtr_function* const func, const char* name, wt_uint32_t argc,
	const wt_typeid* argv, wt_uint32_t retc, const wt_typeid* retv,
	wt_bool pure)
{
	if(NULL==func || NULL==name)
	{
		return WTE_BADARG;
	}
	else
	{
		struct wtr_sym_func looking_for =
		{
			/* sym */
			{
				WTR_SYMTYPE_WTFUNC,
				WTR_SYMBIND_EXPORT,
				NULL,
				name
			},
			/* sig */
			{
				argc,
				retc,
				(wt_typeid*)argv,
				(wt_typeid*)retv,
				(wt_uint8_t)(!!pure)
			},
			/* func_id */
			WTR_SYM_UNDEF
		};

		struct wtr_sym_func* const sym =
			(struct wtr_sym_func*)wtr_vm_find_sym(vm,
				(struct wtr_sym*)&looking_for);

		if(NULL==sym)
		{
			return WTE_NOENT;
		}
		else
		{
			struct wtr_symtab* const symtab = sym->sym.symtab;

			if(NULL!=symtab && 0==(++symtab->refcount))
			{
				--symtab->refcount;
				return WTE_NREF;
			}

			func->func    = sym;
			func->func_id = sym->func_id;
		}
	}

	return WTE_OK;
}

wt_error wtrVMReleaseFunction(struct wtr_vm* const vm,
	struct wtr_function* const func)
{
	if(NULL==func || NULL==func->func)
	{
		return WTE_BADARG;
	}

	if(((struct wtr_sym*)func->func)!=wtr_gidtree_get(&vm->gidtree,
		func->func_id))
	{
		return WTE_BADARG;
	}

	{
		struct wtr_symtab* const symtab = func->func->sym.symtab;

		if(NULL!=symtab)
		{
			--symtab->refcount;
		}
	}

	func->func    = NULL;
	func->func_id = WTR_SYM_UNDEF;

	return WTE_OK;
}

wt_error wtrVMLoadProgram(struct wtr_vm* const vm, wt_program* program,
	struct wtr_context* const ctx, const void* data, size_t size,
	wt_u32 flags)
{
	if(NULL==program || NULL==data || 0==size || NULL==ctx)
	{
		return WTE_BADARG;
	}

	if(WTRR_INIT!=ctx->run)
	{
		return WTE_BADCTX;
	}

	return wtr_loader_sys(vm, program, ctx, data, size, flags);
}

wt_error wtrVMLoadProgramVM(struct wtr_vm* const vm, wt_program* program,
	struct wtr_context* const ctx, wt_vmptr data, size_t size,
	wt_u32 flags)
{
	if(NULL==program || 0==data || 0==size || NULL==ctx)
	{
		return WTE_BADARG;
	}

	if(WTRR_INIT!=ctx->run)
	{
		return WTE_BADCTX;
	}

	return wtr_loader_vm(vm, program, ctx, data, size, flags);
}

wt_error wtrVMReleaseProgram(struct wtr_vm* const vm, wt_program program,
	struct wtr_context* const ctx)
{
	if(NULL==vm || NULL==program || NULL==ctx || vm!=program->vm)
	{
		return WTE_BADARG;
	}

	if(WTRR_INIT!=ctx->run)
	{
		return WTE_BADCTX;
	}

	if(0<program->symtab->refcount)
	{
		return WTE_REFER;
	}

	if(1<program->refcount)
	{
		return WTE_REFER;
	}
	else
	if(0==program->refcount)
	{
		return WTE_DESTRP;
	}

	--program->refcount;

	if(program->pc_dtor)
	{
		wtr_ctx_invoke_dtor(ctx, program);
	}
	else
	{
		wtr_program_destroy(vm, program);
	}

	return WTE_OK;
}

wt_error wtrVMSetUserPointer(struct wtr_vm* const vm, void* ptr)
{
	vm->user_ptr = ptr;
	return WTE_OK;
}

wt_error wtrVMGetUserPointer(struct wtr_vm* const vm, void** pptr)
{
	*pptr = vm->user_ptr;
	return WTE_OK;
}

int wtr_vma_memcpy(struct wtr_vm* const vm, wt_vmptr dst, wt_vmptr src,
	wt_u32 size)
{
	wt_u8 buf[32];

	while(sizeof(buf)<=size)
	{
		if(0!=vmmu32_read(&vm->vmmu, src, buf, sizeof(buf)))
		{
			return 1;
		}

		if(0!=vmmu32_write(&vm->vmmu, dst, buf, sizeof(buf)))
		{
			return 2;
		}

		dst  += sizeof(buf);
		src  += sizeof(buf);
		size -= sizeof(buf);
	}

	if(0<size)
	{
		if(0!=vmmu32_read(&vm->vmmu, src, buf, size))
		{
			return 3;
		}

		if(0!=vmmu32_write(&vm->vmmu, dst, buf, size))
		{
			return 4;
		}
	}

	return 0;
}

int wtr_vma_memmove(struct wtr_vm* const vm, wt_vmptr dst, wt_vmptr src,
	wt_u32 size)
{
	wt_u8 buf[32];

	if(dst<src)
	{
		return wtr_vma_memcpy(vm, dst, src, size);
	}

	src += size;
	dst += size;

	while(sizeof(buf)<=size)
	{
		dst -= sizeof(buf);
		src -= sizeof(buf);

		if(0!=vmmu32_read(&vm->vmmu, src, buf, sizeof(buf)))
		{
			return 1;
		}

		if(0!=vmmu32_write(&vm->vmmu, dst, buf, sizeof(buf)))
		{
			return 2;
		}

		size -= sizeof(buf);
	}

	if(0<size)
	{
		dst -= size;
		src -= size;

		if(0!=vmmu32_read(&vm->vmmu, src, buf, size))
		{
			return 3;
		}

		if(0!=vmmu32_write(&vm->vmmu, dst, buf, size))
		{
			return 4;
		}
	}

	return 0;
}

int wtr_vma_memset(struct wtr_vm* const vm, wt_vmptr vma, int c,
	wt_u32 size)
{
	wt_u8 buf[32];

	memset(buf, c, sizeof(buf));

	while(sizeof(buf)<=size)
	{
		if(0!=vmmu32_write(&vm->vmmu, vma, buf, sizeof(buf)))
		{
			return 1;
		}

		vma  += sizeof(buf);
		size -= sizeof(buf);
	}

	if(0<size)
	{
		if(0!=vmmu32_write(&vm->vmmu, vma, buf, size))
		{
			return 4;
		}
	}

	return 0;
}

int wtr_vma_memcmp(struct wtr_vm* const vm, wt_vmptr s1, wt_vmptr s2,
	wt_u32 size)
{
	wt_u8 buf1[16];
	wt_u8 buf2[sizeof(buf1)];

	while(sizeof(buf1)<=size)
	{
		int res;

		if(0!=vmmu32_read(&vm->vmmu, s1, buf1, sizeof(buf1)))
		{
			return 1;
		}

		if(0!=vmmu32_read(&vm->vmmu, s2, buf2, sizeof(buf2)))
		{
			return 2;
		}

		if(0!=(res=memcmp(buf1, buf2, sizeof(buf1))))
		{
			return res;
		}

		s1   += sizeof(buf1);
		s2   += sizeof(buf1);
		size -= sizeof(buf1);
	}

	if(0<size)
	{
		if(0!=vmmu32_read(&vm->vmmu, s1, buf1, size))
		{
			return 3;
		}

		if(0!=vmmu32_read(&vm->vmmu, s2, buf2, size))
		{
			return 4;
		}

		return memcmp(buf1, buf2, size);
	}

	return 0;
}

int wtr_vma_strlen(struct wtr_vm* const vm, wt_vmptr str, wt_u32* const len)
{
	const wt_vmptr base = str;

	wt_u8 ch;

	do
	{
		if(0!=wtr_vma_read(vm, str++, &ch, sizeof(ch)))
		{
			return 1;
		}
	}
	while(0!=ch);

	*len = (str-1) - base;

	return 0;
}

wt_error wtrVMReadMemory(struct wtr_vm* const vm, wt_vmptr vma,
	void* ptr, wt_u32 size)
{
	if(0!=vmmu32_read(&vm->vmmu, vma, ptr, size))
	{
		return WTE_FAULT;
	}

	return WTE_OK;
}

wt_error wtrVMWriteMemory(struct wtr_vm* const vm, wt_vmptr vma,
	const void* ptr, wt_u32 size)
{
	if(0!=vmmu32_write(&vm->vmmu, vma, ptr, size))
	{
		return WTE_FAULT;
	}

	return WTE_OK;
}
